#include "stdio.h"
#include "rand.cu"

#define POP_SIZE 4096
#define NUM_POINTS 100
#define NUM_GENERATIONS 1000
#define MUTATE 5

struct Point {
    float   x;
    float   y;
    __device__ float distance(float x2, float y2)
    {
	 return sqrt((x-x2)*(x-x2) + (y-y2)*(y-y2));
    }
};

__constant__ Point const_points[NUM_POINTS];

__device__ inline int __abs(int x)
{
        return (x > 0) ? x : -x;
}

// Function prototypes
void savePoints(Point points[]);
void saveSolutions(int population[][NUM_POINTS], float fitness[]);
__device__ __host__ void swap(int *val1, int *val2);
__device__ void crossover(int parent1,int parent2,int child, 
		int *population, int *tempPopulation);
	       //int population[][NUM_POINTS],
	       //int tempPopulation[][NUM_POINTS]);

__global__ void fitness_kernel(int *population, float *fitness)
{
	int tid = threadIdx.x + (blockIdx.x * blockDim.x);
	while (tid < POP_SIZE)
	{
  		float total = 0;
		for (int p = 1; p < NUM_POINTS; p++)
  		{
			int pop1 = *(population + 
					(NUM_POINTS*tid) +
					(p-1));
			int pop2 = *(population + 
					(NUM_POINTS*tid) +
					(p));
			Point p1 = const_points[pop1];	
			Point p2 = const_points[pop2];	
			total += p1.distance(p2.x,p2.y);
  		}
  		fitness[tid] = total;
		tid += blockDim.x * gridDim.x;
	}
}

__global__ void breed_kernel(int *population, float *fitness, int *tempPop)
{
	int tid = threadIdx.x + (blockIdx.x * blockDim.x);
	int parent1, parent2;
	int r1, r2, r3;
	float min;
	r1= __abs(randInt()) % POP_SIZE;
	r2= __abs(randInt()) % POP_SIZE;
	r3= __abs(randInt()) % POP_SIZE;
	min =  fitness[r1];
 	parent1 = r1;
	if (min > fitness[r2])
	{
		min = fitness[r2];
	 	parent1 = r2;
	}
	if (min > fitness[r3])
	{
		min = fitness[r3];
	  	parent1 = r3;
	}
	r1= __abs(randInt()) % POP_SIZE;
	r2= __abs(randInt()) % POP_SIZE;
	r3= __abs(randInt()) % POP_SIZE;
	min =  fitness[r1];
	parent2 = r1;
	if (min > fitness[r2])
	{
	 	min = fitness[r2];
	 	parent2 = r2;
	}
	if (min > fitness[r3])
	{
	 	min = fitness[r3];
	 	parent2 = r3;
	}
	crossover(parent1,parent2,tid,population,tempPop);
	// Mutate with a small chance
	int randMutate = randInt() % 100;
	if (randMutate < MUTATE)
	{
	  int p1 = randInt() % NUM_POINTS;
	  int p2 = randInt() % NUM_POINTS;
	  int index1 = tid*NUM_POINTS + p1;
	  int index2 = tid*NUM_POINTS + p2;
	  swap(&tempPop[index1],&tempPop[index2]);
//	  swap(&tempPopulation[child][p1],
//	     &tempPopulation[child][p2]);
 	}
}

int main()
{
	Point points[NUM_POINTS];
	int *dev_population;
	float *dev_fitness;
	int *dev_tempPop;

	cudaMalloc((void **)&dev_population, sizeof(int)*POP_SIZE*NUM_POINTS);
	cudaMalloc((void **)&dev_tempPop, sizeof(int)*POP_SIZE*NUM_POINTS);
	cudaMalloc((void **)&dev_fitness, sizeof(float)*POP_SIZE);

	// ****** Generate random points
	srand(time(NULL));		// Seed random number generator
	for (int i=0; i<NUM_POINTS;i++)
	{
		points[i].x = ((float) rand() * 10) / RAND_MAX;
		points[i].y = ((float) rand() * 10) / RAND_MAX;
	}
	savePoints(points);		// Save points to a file

	cudaMemcpyToSymbol(const_points, points, sizeof(Point)*NUM_POINTS);

	// ****** Generate random initial population
	int population[POP_SIZE][NUM_POINTS];
	for (int i = 0; i < POP_SIZE; i++)
	{
		// Each individual index 0,1,2,3...NUM_POINTS-1
		for (int j = 0; j < NUM_POINTS; j++)
		{
			population[i][j]=j;
		}
		// Randomize by picking two and swapping them
		for (int j = 0; j < NUM_POINTS; j++)
		{
			int k = rand() % NUM_POINTS;
			swap(&population[i][k],&population[i][j]);
		}
	}

	// Main GA Loop
	float fitness[POP_SIZE];
	int tempPopulation[POP_SIZE][NUM_POINTS];

	for (int gen = 0; gen < NUM_GENERATIONS; gen++)
	{
		float totalFitness = 0;
		float minFitness;
		int best=0;

		cudaMemcpy(dev_population, population, sizeof(int)*POP_SIZE*NUM_POINTS, cudaMemcpyHostToDevice);
		fitness_kernel<<<128,128>>>(dev_population, dev_fitness);
		cudaMemcpy(fitness, dev_fitness, sizeof(float)*POP_SIZE, cudaMemcpyDeviceToHost);

		// Fitness function for each individual
		for (int i = 0; i <POP_SIZE; i++)
		{
			totalFitness += fitness[i];
			if (i==0) 
				minFitness = fitness[i];
			else
			{
				if (fitness[i] < minFitness)
				{
					minFitness = fitness[i];
					best = i;
				}
			}
		}
		printf("Gen %d Best fitness = %f\n",gen, minFitness);

		breed_kernel<<<128,32>>>(dev_population, dev_fitness, 
				dev_tempPop);
		cudaMemcpy(tempPopulation, dev_tempPop, sizeof(int)*POP_SIZE*NUM_POINTS, cudaMemcpyDeviceToHost);

		// Elitism
		memcpy(&tempPopulation[0],&population[best],
			NUM_POINTS*sizeof(int));

		// Copy child pop back to the general population
		memcpy(population,tempPopulation,
			POP_SIZE*NUM_POINTS*sizeof(int));
	}
	saveSolutions(population,fitness);

	cudaFree(dev_population);
	cudaFree(dev_fitness);
	cudaFree(dev_tempPop);
	return 0;
}


void savePoints(Point points[])
{
 FILE *f;
 f = fopen("points.data","w");
 fprintf(f,"%d\n",NUM_POINTS);
 for (int i =0; i < NUM_POINTS; i++)
 {
	fprintf(f,"%f\t%f\n",points[i].x,points[i].y);
 }
 fprintf(f,"-999\n");
 fclose(f);
}

void saveSolutions(int population[][NUM_POINTS],float fitness[])
{
 FILE *f;
 f = fopen("solutions.data","w");
 fprintf(f,"%d\n",POP_SIZE);
 for (int i=0; i<POP_SIZE;i++)
 {
	for (int j=0; j<NUM_POINTS;j++)
		fprintf(f,"%d ",population[i][j]);
	fprintf(f," %f\n",fitness[i]);
 }
 fprintf(f,"-999\n");
 fclose(f);
}

__device__ __host__ void swap(int *val1, int *val2)
{
  int temp = *val1;
  *val1 = *val2;
  *val2 = temp;
}


// Copy random segment from one parent, then points in order
// from the second parent if not already in the child
__device__ void crossover(int parent1,int parent2,int child, 
		int *population, int *tempPop)
//	       int population[][NUM_POINTS],
//	       int tempPopulation[][NUM_POINTS])
{
 char copied[NUM_POINTS];
 for (int i = 0; i < NUM_POINTS; i++)
	copied[i]=0;
 int start, end;
 start = __abs(randInt()) % NUM_POINTS;
 end = __abs(randInt()) % NUM_POINTS;
 if (start > end)
	swap(&start,&end);
 // Copy from start..end from parent1
 for (int i = start; i <= end; i++)
 {
	int parValue = *(population + 
				(NUM_POINTS*parent1) +
				(i));
	int tempIndex = ((NUM_POINTS*child) + i);
	tempPop[tempIndex] = parValue;
	copied[parValue] = 1;
	//	tempPopulation[child][i]=population[parent1][i];
	//	copied[population[parent1][i]]=1;
 }
 // Copy everything else from parent2
 int j = 0;
 for (int i = 0; i < NUM_POINTS; i++)
 {
 	if (j==start) j=end+1;
	if (j == NUM_POINTS)
	{
		break;
	}
	int parValue = *(population + 
				(NUM_POINTS*parent2) +
				(i));
	int tempIndex = ((NUM_POINTS*child) + j);
	if (!copied[parValue])
	{
		tempPop[tempIndex] = parValue;
		j++;
	}
/*	if (!copied[population[parent2][i]])
	{
		tempPopulation[child][j] = population[parent2][i];
		j++;
	}
*/
 }
}

